"""
The `~certbot_dns_azure.dns_azure` plugin automates the process of
completing a ``dns-01`` challenge (`~acme.challenges.DNS01`) by creating, and
subsequently removing, TXT records using the Azure API.

.. note::
   The plugin is not installed by default. It can be installed by heading to
   `certbot.eff.org <https://certbot.eff.org/instructions#wildcard>`_, choosing your system and
   selecting the Wildcard tab.

Named Arguments
---------------

========================================  =====================================
``--dns-azure-config``                    Azure config INI file.
                                          (Required)
``--dns-azure-credentials``               Alias for --dns-azure-config
========================================  =====================================


Configuration
-------------

Use of this plugin requires a configuration file containing Azure API
credentials or information.

This plugin supported API authentication using either Service Principals or
utilising a Managed Identity assigned to the virtual machine.

Regardless which authentication method used, the identity will need the
"DNS Zone Contributor" role assigned to it.

As multiple Azure DNS Zones in multiple resource groups can exist, the config
file needs a mapping of zone to resource group ID. Multiple zones -> ID mappings
can be listed by using the key ``dns_azure_zoneX`` where X is a unique number.
At least 1 zone mapping is required.

.. code-block:: ini
   :name: certbot_azure_service_principal.ini
   :caption: Example config file using a service principal

   dns_azure_sp_client_id = 912ce44a-0156-4669-ae22-c16a17d34ca5
   dns_azure_sp_client_secret = E-xqXU83Y-jzTI6xe9fs2YC~mck3ZzUih9
   dns_azure_tenant_id = ed1090f3-ab18-4b12-816c-599af8a88cf7

   dns_azure_environment = "AzurePublicCloud"

   dns_azure_zone1 = example.com:/subscriptions/c135abce-d87d-48df-936c-15596c6968a5/resourceGroups/dns1
   dns_azure_zone2 = example.org:/subscriptions/99800903-fb14-4992-9aff-12eaf2744622/resourceGroups/dns2

.. code-block:: ini
   :name: certbot_azure_service_principal_certificate.ini
   :caption: Example config file using a service principal with certificate

   dns_azure_sp_client_id = 912ce44a-0156-4669-ae22-c16a17d34ca5
   dns_azure_sp_certificate_path = /path/to/certificate.pem
   dns_azure_tenant_id = ed1090f3-ab18-4b12-816c-599af8a88cf7

   dns_azure_environment = "AzurePublicCloud"

   dns_azure_zone1 = example.com:/subscriptions/c135abce-d87d-48df-936c-15596c6968a5/resourceGroups/dns1
   dns_azure_zone2 = example.org:/subscriptions/99800903-fb14-4992-9aff-12eaf2744622/resourceGroups/dns2

.. code-block:: ini
   :name: certbot_azure_user_msi.ini
   :caption: Example config file using user assigned MSI:

   dns_azure_msi_client_id = 912ce44a-0156-4669-ae22-c16a17d34ca5

   dns_azure_zone1 = example.com:/subscriptions/c135abce-d87d-48df-936c-15596c6968a5/resourceGroups/dns1
   dns_azure_zone2 = example.org:/subscriptions/99800903-fb14-4992-9aff-12eaf2744622/resourceGroups/dns2

.. code-block:: ini
   :name: certbot_azure_system_msi.ini
   :caption: Example config file using system assigned MSI:

   dns_azure_msi_system_assigned = true

   dns_azure_environment = "AzurePublicCloud"

   dns_azure_zone1 = example.com:/subscriptions/c135abce-d87d-48df-936c-15596c6968a5/resourceGroups/dns1
   dns_azure_zone2 = example.org:/subscriptions/99800903-fb14-4992-9aff-12eaf2744622/resourceGroups/dns2

.. code-block:: ini
   :name: certbot_azure_cli.ini
   :caption: Example config file using Azure CLI credentials:

   dns_azure_use_cli_credentials = true

   dns_azure_environment = "AzurePublicCloud"

   dns_azure_zone1 = example.com:/subscriptions/c135abce-d87d-48df-936c-15596c6968a5/resourceGroups/dns1
   dns_azure_zone2 = example.org:/subscriptions/99800903-fb14-4992-9aff-12eaf2744622/resourceGroups/dns2

.. code-block:: ini
   :name: certbot_azure_workload_identity.ini
   :caption: Example config file using Azure Workload Identity credentials:

   dns_azure_use_workload_identity_credentials = true

   dns_azure_environment = "AzurePublicCloud"

   dns_azure_zone1 = example.com:/subscriptions/c135abce-d87d-48df-936c-15596c6968a5/resourceGroups/dns1
   dns_azure_zone2 = example.org:/subscriptions/99800903-fb14-4992-9aff-12eaf2744622/resourceGroups/dns2


The path to this file can be provided interactively or using the
``--dns-azure-config`` command-line argument. Certbot records the path
to this file for use during renewal, but does not store the file's contents.

.. caution::
   You should protect these API credentials as you would the password to your
   Azure account. Users who can read this file can use these credentials
   to issue arbitrary API calls on your behalf. Users who can cause Certbot to
   run using these credentials can complete a ``dns-01`` challenge to acquire
   new certificates or revoke existing certificates for domains the identity
   has access to.

Certbot will emit a warning if it detects that the credentials file can be
accessed by other users on your system. The warning reads "Unsafe permissions
on configuration file", followed by the path to the config
file. This warning will be emitted each time Certbot uses the config file,
including for renewal, and cannot be silenced except by addressing the issue
(e.g., by using a command like ``chmod 600`` to restrict access to the file).


Azure Environment
-----------------

The Azure Cloud will default to ``AzurePublicCloud``, this is used to change
the authentication endpoint used when generating credentials. This option
can be specified in the config file using ``dns_azure_environment`` or 
via an environment variable ``AZURE_ENVIRONMENT``.

The supported values are:

========================================  =====================================
``AzurePublicCloud``                      https://management.azure.com/
``AzureUSGovernmentCloud``                https://management.usgovcloudapi.net/
``AzureChinaCloud``                       https://management.chinacloudapi.cn/
``AzureGermanCloud``                      https://management.microsoftazure.de/
========================================  =====================================

DNS delegation
--------------

DNS delegation, also known as DNS aliasing, is a process of allowing a secondary DNS zone to handle validation in place
of the primary zone. For example, you would like to acquire a certificate for ``example.com`` but have the validation
performed on a secondary domain ``example.org``. You would create a ``_acme-challenge.example.com`` CNAME on the
``example.com`` nameserver with the value of ``_acme-challenge.example.org``. The ACME server will resolve the CNAME and
validate the TXT record ``_acme-challenge.example.org`` instead. Certbot itself does not support CNAME aliasing,
therefore this plugin does what it can to support it.

The common reasons for DNS delegation are:
 * The primary DNS zone is hosted on a nameserver with no API access
 * Security concerns regarding access to the primary DNS zone

We'll use two domains for the examples below: ``foo.com`` and ``bar.com``.

Example: Primary Zone, no API Access
++++++++++++++++++++++++++++++++++++

Let's assume you wish to get a certificate for ``test.foo.com``, this will result in a validation record for
``_acme-challenge.test.foo.com``. Assuming you don't have API access to the ``foo.com`` zone, you can manually CNAME said
validation record to point it to ``bar.com``. E.g.
``_acme-challenge.test.foo.com CNAME -> _acme-challenge.test.foo.com.bar.com`` This will result in a TXT record called
``_acme-challenge.test.foo.com`` created in the ``bar.com`` zone.

This can be achieved by using the following config snippet

.. code-block:: ini
    :name: certbot_azure_alias1.ini
    :caption: Example configuration snippet for DNS delegation

    dns_azure_zone1 = test.foo.com:/subscriptions/c135abce-d87d-48df-936c-15596c6968a/resourceGroups/dns1/providers/Microsoft.Network/dnszones/bar.com

So when the Azure Certbot plugin gets requested to make ``_acme-challenge.test.foo.com``, the zone that record gets
created in is overridden to ``bar.com`` hence the entire ``_acme-challenge.test.foo.com`` is needed as the prefix in the
CNAME value before ``bar.com``.

Example: Delegation + more security
+++++++++++++++++++++++++++++++++++

One can go a step further than the step above when hosting records in Azure. Instead of granting Certbot write access
to an entire DNS Zone, you can grant access to specific records.

As with before, we shall get a certificate for ``test.foo.com``. We shall make a CNAME record like:
``_acme-challenge.test.foo.com CNAME test_validation.bar.com`` (yes it doesn't *need* ``_acme-challenge`` in this example).

Using a config like the following:

.. code-block:: ini
    :name: certbot_azure_alias2.ini
    :caption: Example configuration snippet for DNS delegation

    dns_azure_zone1 = test.foo.com:/subscriptions/c135abce-d87d-48df-936c-15596c6968a/resourceGroups/dns1/providers/Microsoft.Network/dnszones/bar.com/TXT/test_validation

This **requires** you to create a TXT record called ``test_validation`` in the ``bar.com`` zone, the value should be ``-`` and certbot should have IAM privileges to
write to that record explicitly. Now when the request is sent to the plugin to create ``_acme-challenge.test.foo.com``
the zone its created in will be overridden with ``bar.com`` and the validation record will be overridden with ``test_validation``. This
effectively does the same thing as if certbot had actually resolved the CNAME (which it doesn't).

Now one caveat here, if you assign specific IAM roles to an individual record, they'll be lost when the record is cleaned up and deleted.
Which means when it comes to renewal, it'll fail as it has no IAM privileges to update said record as it no longer exists. For this reason
when the config has record ID's in it, it will not delete the validation records, it will just set its value to ``-``
(hence you were told to set this initially).

This example can be simplified to allow you to reduce the certbot permissions but **not** do CNAME delegation, so
for ``test.foo.com`` you would make a TXT record for ``_acme-challenge.test.foo.com`` with the value of ``-`` and then use a config
snippet like:

.. code-block:: ini
    :name: certbot_azure_moresecure.ini
    :caption: Example configuration snippet for individual record permissions

    dns_azure_zone1 = test.foo.com:/subscriptions/c135abce-d87d-48df-936c-15596c6968a/resourceGroups/dns1/providers/Microsoft.Network/dnszones/foo.com/TXT/_acme-validation.test

This will override the zone to foo.com (which ``test`` is already in) and the validation record (though it's overridden to the same thing) but now
it will not delete said validation record.


Generic Certbot Examples
------------------------

.. code-block:: bash
   :caption: To acquire a certificate for ``example.com``

   certbot certonly \\
     --dns-azure-config ~/.secrets/certbot/azure.ini \\
     -d example.com

.. code-block:: bash
   :caption: To acquire a single certificate for both ``example.com`` and
             ``example.org``

   certbot certonly \\
     --dns-azure-config ~/.secrets/certbot/azure.ini \\
     -d example.com \\
     -d example.org

To run in a non-interactive manner:

.. code-block:: bash
   :caption: Non-interactive
   
   certbot certonly \\
     --authenticator dns-azure \\
     --preferred-challenges dns \\
     --noninteractive \\
     --agree-tos \\
     --dns-azure-config ~/.secrets/certbot/azure.ini \\
     -d example.com

"""



"""DNS Authenticator for Azure DNS."""
import logging
import time
import random
from os import getenv
from typing import Dict, Tuple

from azure.mgmt.dns import DnsManagementClient
from azure.mgmt.dns.models import RecordSet, TxtRecord
from azure.core.exceptions import HttpResponseError
from azure.core.utils import CaseInsensitiveDict
from azure.identity import ClientSecretCredential, ManagedIdentityCredential, CertificateCredential, AzureCliCredential, WorkloadIdentityCredential

from certbot import errors
from certbot.plugins import dns_common

logger = logging.getLogger(__name__)
logging.getLogger('azure').setLevel(logging.WARNING)


class Authenticator(dns_common.DNSAuthenticator):
    """DNS Authenticator for Azure DNS

    This Authenticator uses the Azure DNS API to fulfill a dns-01 challenge.
    """

    description = ('Obtain certificates using a DNS TXT record (if you are using '
                   'Azure for DNS).')
    ttl = 120

    def __init__(self, *args, **kwargs):
        super(Authenticator, self).__init__(*args, **kwargs)
        self.credential = None
        self.domain_zoneid = {}  # type: Dict[str, str]

        # Azure Environmental Support
        self._azure_environment = getenv("AZURE_ENVIRONMENT", "AzurePublicCloud").lower()
        self._azure_endpoints = {
            "azurepubliccloud": {
                "ResourceManagerEndpoint": "https://management.azure.com/",
                "ActiveDirectoryEndpoint": "https://login.microsoftonline.com/"
            },
            "azureusgovernmentcloud": {
                "ResourceManagerEndpoint": "https://management.usgovcloudapi.net/",
                "ActiveDirectoryEndpoint": "https://login.microsoftonline.us/"
            },
            "azurechinacloud": {
                "ResourceManagerEndpoint": "https://management.chinacloudapi.cn/",
                "ActiveDirectoryEndpoint": "https://login.chinacloudapi.cn/"
            },
            "azuregermancloud": {
                "ResourceManagerEndpoint": "https://management.microsoftazure.de/",
                "ActiveDirectoryEndpoint": "https://login.microsoftonline.de/"
            }
        }

    @classmethod
    def add_parser_arguments(cls, add):  # pylint: disable=arguments-differ
        super(Authenticator, cls).add_parser_arguments(add)
        add('config', help='Azure config INI file.')
        add('credentials', help='Azure config INI file. Fallback for legacy integrations')

    def more_info(self):  # pylint: disable=missing-function-docstring
        return 'This plugin configures a DNS TXT record to respond to a dns-01 challenge using ' + \
               'the Azure DNS API.'

    def _validate_credentials(self, credentials):
        sp_client_id = credentials.conf('sp_client_id')
        sp_client_secret = credentials.conf('sp_client_secret')
        sp_certificate_path = credentials.conf('sp_certificate_path')
        tenant_id = credentials.conf('tenant_id')
        has_sp = all((sp_client_id, any((sp_client_secret, sp_certificate_path)), tenant_id))

        msi_client_id = credentials.conf('msi_client_id')
        msi_system_assigned = credentials.conf('msi_system_assigned')

        use_azure_cli_creds = credentials.conf('use_cli_credentials')

        use_workload_identity_creds = credentials.conf('use_workload_identity_credentials')

        if not any((has_sp, msi_system_assigned, msi_client_id, use_azure_cli_creds, use_workload_identity_creds)):
            raise errors.PluginError('{}: No authentication methods have been '
                                     'configured for Azure DNS. Either configure '
                                     'a service principal, system/user assigned '
                                     'managed identity or configure the use of '
                                     'azure cli or workload identity credentials'.format(credentials.confobj.filename))

        has_zone_mapping = any((key for key in credentials.confobj.keys() if 'azure_zone' in key))

        if not has_zone_mapping:
            raise errors.PluginError('{}: At least one zone mapping needs to be provided,'
                                     ' e.g dns_azure_zone1 = DOMAIN:DNS_ZONE_RESOURCE_GROUP_ID'
                                     ''.format(credentials.confobj.filename))

        # Azure Environment
        environment = credentials.conf('environment')

        if environment:
            self._azure_environment = environment.lower()

        self._arm_endpoint = self._azure_endpoints[self._azure_environment]["ResourceManagerEndpoint"]
        self._aad_endpoint = self._azure_endpoints[self._azure_environment]["ActiveDirectoryEndpoint"]
        
        # Check we have key value
        dns_zone_mapping_items_has_colon = [':' in value
                                            for key, value in credentials.confobj.items()
                                            if 'azure_zone' in key]
        if not all(dns_zone_mapping_items_has_colon):
            raise errors.PluginError('{}: DNS Zone mapping is not in the format of '
                                     'DOMAIN:DNS_ZONE_RESOURCE_GROUP_ID'
                                     ''.format(credentials.confobj.filename))

    def _setup_credentials(self):
        # Alias's dns-azure-credentials -> dns-azure-config
        if self.config.namespace.dns_azure_credentials:
            self.config.namespace.dns_azure_config = self.config.namespace.dns_azure_credentials

        valid_creds = self._configure_credentials(
            'config',
            'Azure config INI file',
            None,
            self._validate_credentials
        )

        # Convert dns_azure_zoneX = key:value into key:value
        dns_zone_mapping_items = [value for key, value in valid_creds.confobj.items()
                                  if 'azure_zone' in key]
        self.domain_zoneid = dict([item.split(':', 1) for item in dns_zone_mapping_items])

        # Figure out which credential type we're going to use
        sp_client_id = valid_creds.conf('sp_client_id')
        sp_client_secret = valid_creds.conf('sp_client_secret')
        sp_certificate_path = valid_creds.conf('sp_certificate_path')
        tenant_id = valid_creds.conf('tenant_id')
        msi_client_id = valid_creds.conf('msi_client_id')
        use_azure_cli_creds = valid_creds.conf('use_cli_credentials')
        use_workload_identity_creds = valid_creds.conf('use_workload_identity_credentials')

        self.credential = self._get_azure_credentials(
            sp_client_id, sp_client_secret, sp_certificate_path, tenant_id, msi_client_id, use_azure_cli_creds, use_workload_identity_creds, self._aad_endpoint
        )

    @staticmethod
    def _get_azure_credentials(client_id=None, client_secret=None, certificate_path=None, tenant_id=None, msi_client_id=None,
                               use_azure_cli_creds=None, use_workload_identity_creds=None, aad_endpoint=None):
        has_sp = all((client_id, client_secret, tenant_id))
        has_sp_cert = all((client_id, certificate_path, tenant_id))
        if use_azure_cli_creds:  # TODO move to DefaultAzureCredential
            return AzureCliCredential(tenant_id=tenant_id)
        elif use_workload_identity_creds:
            return WorkloadIdentityCredential(tenant_id=tenant_id)
        elif has_sp:
            return ClientSecretCredential(
                client_id=client_id,
                client_secret=client_secret,
                tenant_id=tenant_id,
                authority=aad_endpoint
            )
        elif has_sp_cert:
            return CertificateCredential(
                client_id=client_id,
                certificate_path=certificate_path,
                tenant_id=tenant_id,
                authority=aad_endpoint
            )
        elif msi_client_id:
            return ManagedIdentityCredential(client_id=msi_client_id)
        else:
            return ManagedIdentityCredential()

    def _get_ids_for_domain(self, domain: str, validation_name: str) -> Tuple[str, str, str, str, bool]:
        """
        :param domain: Domain/subdomain to look up the closest parent in the config file
        :param validation_name: DNS challenge record name, fully qualified

        This returns:
        * The Azure DNS zone for which to add records to
        * The subscription ID for said zone
        * The resource group for said zone
        * The relative validation record name (or if explicitly overrided with an ID, an alternate record name)
        * If the validation record can be deleted, if its explicitly overrided, it wont be deleted but set to `-`
        """
        # So if the config contains domain.io and test.domain.io
        # and we want to renew, we'd prefer test.domain.io.
        # Sort domains by longest first and then attempt to find the right one.
        # This should work better, as then a.b.test.domain.io would pick domain.io irrelevant
        # of its order in the config
        azure_domains = sorted(self.domain_zoneid.keys(), key=lambda domain: len(domain), reverse=True)

        try:
            for azure_dns_domain in azure_domains:
                # Look to see if domain ends with key, to cover subdomains
                if domain.endswith(azure_dns_domain):
                    zone_id = self.domain_zoneid[azure_dns_domain]

                    try:
                        resource = self.parse_azure_resource_id(zone_id)
                    except ValueError as exc:
                        raise errors.PluginError('Failed to parse resource ID for {}: {}'
                                                 .format(domain, zone_id)) from exc
                    subscription_id = resource.get('subscriptions')
                    rg_name = resource.get('resourceGroups')
                    if 'dnsZones' in resource:  # If we're manually specifying an alternate zone to use, override.
                        azure_dns_domain = resource.get('dnsZones')
                    relative_validation_name = self._get_relative_domain(validation_name, azure_dns_domain)
                    can_delete = True
                    if 'TXT' in resource:  # If we're explicitly specifing a destination record, use instead.
                        relative_validation_name = resource.get('TXT')
                        can_delete = False  # If we're specifying a specific record, dont delete it

                    return azure_dns_domain, subscription_id, rg_name, relative_validation_name, can_delete
            else:
                raise errors.PluginError('Domain {} does not have a valid domain to '
                                         'resource group id mapping'.format(domain))
        except IndexError:
            raise errors.PluginError('Domain {} has an invalid resource group id'.format(domain))

    @staticmethod
    def _get_relative_domain(fqdn: str, domain: str) -> str:
        if fqdn == domain:
            return '@'
        return fqdn.replace(domain, '').strip('.')

    def _perform(self, domain, validation_name, validation, retry_attempt=0):
        azure_domain, subscription_id, resource_group_name, validation_name, _ = self._get_ids_for_domain(domain, validation_name)
        client = self._get_azure_client(subscription_id)

        # Check to see if there are any existing TXT validation record values
        txt_value = {validation}
        etag = None
        try:
            existing_rr = client.record_sets.get(
                resource_group_name=resource_group_name,
                zone_name=azure_domain,
                relative_record_set_name=validation_name,
                record_type='TXT')
            etag = existing_rr.etag
            for record in existing_rr.txt_records:
                for value in record.value:
                    if value == '-':
                        continue
                    txt_value.add(value)
        except HttpResponseError as err:
            if err.status_code != 404:  # Ignore RR not found
                raise errors.PluginError('Failed to check TXT record for domain '
                                         '{}, error: {}'.format(domain, err))

        try:
            client.record_sets.create_or_update(
                resource_group_name=resource_group_name,
                zone_name=azure_domain,
                relative_record_set_name=validation_name,
                record_type='TXT',
                if_match=etag,
                parameters=RecordSet(ttl=self.ttl, txt_records=[TxtRecord(value=[v]) for v in txt_value])
            )
        except HttpResponseError as err:
            if err.status_code == 412:
                # There is some parallel access on this record, sleep a random amount and try again.
                if retry_attempt > 10:
                    raise errors.PluginError('Failed to add TXT record for domain {}, max retries due to concurrent access exceeded'
                                             ', error: {}'.format(domain, err))
                sleep_secs = random.randint(1, 10)
                retry_attempt += 1
                logger.warning("Concurrent access to record {}, sleeping {} seconds, retry attempt: {}".format(domain, sleep_secs, retry_attempt))
                time.sleep(sleep_secs)
                self._perform(domain, validation_name, validation, retry_attempt)
            else:
                raise errors.PluginError('Failed to add TXT record to domain '
                                         '{}, error: {}'.format(domain, err))

    def _cleanup(self, domain, validation_name, validation, retry_attempt=0):
        if self.credential is None:
            self._setup_credentials()

        azure_domain, subscription_id, resource_group_name, validation_name, can_delete = self._get_ids_for_domain(domain, validation_name)
        client = self._get_azure_client(subscription_id)

        txt_value = set()
        etag = None
        try:
            existing_rr = client.record_sets.get(resource_group_name=resource_group_name,
                                                 zone_name=azure_domain,
                                                 relative_record_set_name=validation_name,
                                                 record_type='TXT')
            etag = existing_rr.etag
            for record in existing_rr.txt_records:
                for value in record.value:
                    txt_value.add(value)
        except HttpResponseError as err:
            if err.status_code != 404:  # Ignore RR not found
                raise errors.PluginError('Failed to check TXT record for domain '
                                         '{}, error: {}'.format(domain, err))

        txt_value -= {validation}

        try:
            if txt_value:
                client.record_sets.create_or_update(
                    resource_group_name=resource_group_name,
                    zone_name=azure_domain,
                    relative_record_set_name=validation_name,
                    record_type='TXT',
                    if_match=etag,
                    parameters=RecordSet(ttl=self.ttl,
                                         txt_records=[TxtRecord(value=[v]) for v in txt_value])
                )
            else:
                if can_delete:
                    client.record_sets.delete(
                        resource_group_name=resource_group_name,
                        zone_name=azure_domain,
                        relative_record_set_name=validation_name,
                        if_match=etag,
                        record_type='TXT'
                    )
                else:
                    client.record_sets.create_or_update(  # We've manually specified a record, so dont delete, set to -
                        resource_group_name=resource_group_name,
                        zone_name=azure_domain,
                        relative_record_set_name=validation_name,
                        if_match=etag,
                        record_type='TXT',
                        parameters=RecordSet(ttl=self.ttl, txt_records=[TxtRecord(value=['-'])])
                    )
        except HttpResponseError as err:
            if err.status_code == 412:
                # There is some parallel access on this record, sleep a random amount and try again.
                if retry_attempt > 10:
                    raise errors.PluginError('Failed to remove/empty TXT record for domain {}, max retries due to concurrent access exceeded'
                                             ', error: {}'.format(domain, err))
                sleep_secs = random.randint(1, 10)
                retry_attempt += 1
                logger.warning("Concurrent access to record {}, sleeping {} seconds, retry attempt: {}".format(domain, sleep_secs, retry_attempt))
                time.sleep(sleep_secs)
                self._cleanup(domain, validation_name, validation, retry_attempt)
            elif err.status_code != 404:  # Ignore RR not found
                raise errors.PluginError('Failed to remove/empty TXT record for domain '
                                         '{}, error: {}'.format(domain, err))

    def _get_azure_client(self, subscription_id):
        """
        Gets azure DNS client

        :param subscription_id: Azure subscription ID
        :type subscription_id: str
        :return: Azure DNS client
        :rtype: DnsManagementClient
        """
        return DnsManagementClient(self.credential, subscription_id, None, self._arm_endpoint, credential_scopes=[self._arm_endpoint + "/.default"])

    @staticmethod
    def parse_azure_resource_id(resource_id):
        rsrc_id = resource_id
        if rsrc_id.startswith('/'):
            rsrc_id = rsrc_id[1:]

        if rsrc_id.endswith('/'):
            rsrc_id = rsrc_id[:-1]

        if '/' not in rsrc_id:
            raise ValueError('Invalid resource ID: {}'.format(resource_id))

        parts = rsrc_id.split('/')
        if (len(parts) % 2) != 0 or '' in parts:
            raise ValueError('Invalid resource ID: {}'.format(resource_id))
        return CaseInsensitiveDict(zip(parts[0::2], parts[1::2]))



"""Tests for certbot_dns_azure._internal.dns_azure."""

import unittest

try:
    import mock
except ImportError: # pragma: no cover
    from unittest import mock # type: ignore

from certbot import errors
from certbot.compat import os
from certbot import achallenges
from certbot.plugins import dns_test_common
from certbot.plugins.dns_test_common import KEY
from certbot.tests import util as test_util, acme_util

from azure.mgmt.dns.models import RecordSet, TxtRecord

MULTI_DOMAIN = [
    achallenges.KeyAuthorizationAnnotatedChallenge(
        challb=acme_util.DNS01, domain='example.com', account_key=KEY),
    achallenges.KeyAuthorizationAnnotatedChallenge(
        challb=acme_util.DNS01, domain='example.org', account_key=KEY),
    achallenges.KeyAuthorizationAnnotatedChallenge(
        challb=acme_util.DNS01, domain='example.net', account_key=KEY)
]
SINGLE_DOMAIN = [
    achallenges.KeyAuthorizationAnnotatedChallenge(
        challb=acme_util.DNS01, domain='example.com', account_key=KEY),
]
SUB_DOMAIN = [
    achallenges.KeyAuthorizationAnnotatedChallenge(
        challb=acme_util.DNS01, domain='a.b.example.com', account_key=KEY),
]


class AuthenticatorTest(test_util.TempDirTestCase, dns_test_common.BaseAuthenticatorTest):

    def setUp(self):
        from certbot_dns_azure._internal.dns_azure import Authenticator

        super(AuthenticatorTest, self).setUp()

        # This causes some weird zope errors
        mock.patch("certbot.display.util.notify", lambda x: ...).start()

        # Setup config files
        config_files = (
            ('sp.ini', {
                'azure_sp_client_id': '912ce44a-0156-4669-ae22-c16a17d34ca5',
                'azure_sp_client_secret': 'E-xqXU83Y-jzTI6xe9fs2YC~mck3ZzUih9',
                'azure_tenant_id': 'ed1090f3-ab18-4b12-816c-599af8a88cf7',
                'azure_zone1': 'example.com:/subscriptions/c135abce-d87d-48df-936c-15596c6968a5/resourceGroups/dns1',
                'azure_zone2': 'example.org:/subscriptions/99800903-fb14-4992-9aff-12eaf2744622/resourceGroups/dns2',
                'azure_zone3': ('example.net:/subscriptions/99800903-fb14-4992-9aff-12eaf2744622/resourceGroups/dns2'
                                '/providers/Microsoft.Network/dnsZones/example.com')
            }),
            ('sp_cert.ini', {
                'azure_sp_client_id': '912ce44a-0156-4669-ae22-c16a17d34ca5',
                'azure_sp_client_secret': 'E-xqXU83Y-jzTI6xe9fs2YC~mck3ZzUih9',
                'azure_certificate_path': '/path/to/cert.pem',
                'azure_zone1': 'example.com:/subscriptions/c135abce-d87d-48df-936c-15596c6968a5/resourceGroups/dns1',
                'azure_zone2': 'example.org:/subscriptions/99800903-fb14-4992-9aff-12eaf2744622/resourceGroups/dns2'
            }),
            ('user_assigned_msi.ini', {
                'azure_msi_client_id': '912ce44a-0156-4669-ae22-c16a17d34ca5',
                'azure_zone1': 'example.com:/subscriptions/c135abce-d87d-48df-936c-15596c6968a5/resourceGroups/dns1',
                'azure_zone2': 'example.org:/subscriptions/99800903-fb14-4992-9aff-12eaf2744622/resourceGroups/dns2'
            }),
            ('system_msi.ini', {
                'azure_msi_system_assigned': 'true',
                'azure_zone1': 'example.com:/subscriptions/c135abce-d87d-48df-936c-15596c6968a5/resourceGroups/dns1',
                'azure_zone2': 'example.org:/subscriptions/99800903-fb14-4992-9aff-12eaf2744622/resourceGroups/dns2'
            })
        )
        for file, config in config_files:
            dns_test_common.write(config, os.path.join(self.tempdir, file))

        self.sp_config = mock.MagicMock(
            azure_config=os.path.join(self.tempdir, 'sp.ini'),
            azure_propagation_seconds=0)
        self.sp_cert_config = mock.MagicMock(
            azure_config=os.path.join(self.tempdir, 'sp_cert.ini'),
            azure_propagation_seconds=0)
        self.umsi_config = mock.MagicMock(
            azure_config=os.path.join(self.tempdir, 'user_assigned_msi.ini'),
            azure_propagation_seconds=0)
        self.smsi_config = mock.MagicMock(
            azure_config=os.path.join(self.tempdir, 'system_msi.ini'),
            azure_propagation_seconds=0)

        self.auth = Authenticator(self.sp_config, "azure")
        self.mock_credentials = mock.MagicMock()
        self.mock_client = mock.MagicMock()
        self.auth._get_azure_credentials = mock.MagicMock(return_value=self.mock_credentials)
        self.auth._get_azure_client = mock.MagicMock(return_value=self.mock_client)

    def test_perform_multidomain(self):
        self.mock_client.record_sets.get.return_value = RecordSet(txt_records=[])

        # Extract zone TXT record name and value
        zone1_req, zone2_req, zone3_req = MULTI_DOMAIN
        zone1_domain_name = zone1_req.validation_domain_name(zone1_req.domain)
        zone1_relative_record = zone1_domain_name.replace('example.com', '').strip('.')
        zone1_key = zone1_req.validation(zone1_req.account_key)
        zone2_domain_name = zone2_req.validation_domain_name(zone2_req.domain)
        zone2_relative_record = zone2_domain_name.replace('example.org', '').strip('.')
        zone2_key = zone2_req.validation(zone2_req.account_key)
        zone3_domain_name = zone3_req.validation_domain_name(zone3_req.domain)
        zone3_relative_record = zone3_domain_name.replace('example.com', '').strip('.')
        zone3_key = zone3_req.validation(zone3_req.account_key)

        self.auth.perform(MULTI_DOMAIN)

        # Check azure client call counts
        self.assertEqual(self.mock_client.record_sets.get.call_count, 3)
        self.assertEqual(self.mock_client.record_sets.create_or_update.call_count, 3)

        #
        zone1_call, zone2_call, zone3_call = self.mock_client.record_sets.create_or_update.call_args_list
        self.assertEqual(zone1_call[1]['zone_name'], "example.com")
        self.assertEqual(zone1_call[1]['record_type'], "TXT")
        self.assertEqual(zone1_call[1]['relative_record_set_name'], zone1_relative_record)
        zone1_txt_records = zone1_call[1]['parameters'].txt_records
        self.assertEqual(len(zone1_txt_records), 1)
        self.assertEqual(zone1_txt_records[0].value[0], zone1_key)

        self.assertEqual(zone2_call[1]['zone_name'], "example.org")
        self.assertEqual(zone2_call[1]['relative_record_set_name'], zone2_relative_record)
        zone2_txt_records = zone2_call[1]['parameters'].txt_records
        self.assertEqual(len(zone2_txt_records), 1)
        self.assertEqual(zone2_txt_records[0].value[0], zone2_key)

        # Test DNS delegation of example.net to example.com
        self.assertEqual(zone3_req.domain, "example.net")
        self.assertEqual(zone3_call[1]['zone_name'], "example.com")
        self.assertEqual(zone3_call[1]['relative_record_set_name'], zone3_relative_record)
        zone3_txt_records = zone3_call[1]['parameters'].txt_records
        self.assertEqual(len(zone3_txt_records), 1)
        self.assertEqual(zone3_txt_records[0].value[0], zone3_key)

    def test_perform_existing(self):
        self.mock_client.record_sets.get.return_value = RecordSet(txt_records=[
            TxtRecord(value=['someexistingkey'])
        ])

        # Extract zone TXT record name and value
        zone1_req = SINGLE_DOMAIN[0]
        zone1_domain_name = zone1_req.validation_domain_name(zone1_req.domain)
        zone1_relative_record = zone1_domain_name.replace('example.com', '').strip('.')
        zone1_key = zone1_req.validation(zone1_req.account_key)

        self.auth.perform(SINGLE_DOMAIN)

        # Check azure client call counts
        self.assertEqual(self.mock_client.record_sets.get.call_count, 1)
        self.assertEqual(self.mock_client.record_sets.create_or_update.call_count, 1)

        #
        expected = [self.mock_client.record_sets.create_or_update.call(
            resource_group_name='dns1',
            zone_name='example.com',
            relative_record_set_name=zone1_domain_name,
            parameters=RecordSet(txt_records=[TxtRecord(value=[zone1_key]), TxtRecord(value=['someexistingkey'])])
        )]
        zone1_call = self.mock_client.record_sets.create_or_update.call_args_list[0]
        self.assertEqual(zone1_call[1]['zone_name'], "example.com")
        self.assertEqual(zone1_call[1]['record_type'], "TXT")
        self.assertEqual(zone1_call[1]['relative_record_set_name'], zone1_relative_record)
        zone1_txt_records = zone1_call[1]['parameters'].txt_records

        self.assertEqual(len(zone1_txt_records), 2)
        txt_values = [rr.value for rr in zone1_txt_records]
        self.assertIn([zone1_key], txt_values)
        self.assertIn(['someexistingkey'], txt_values)

    def test_perform_subdomain(self):
        self.mock_client.record_sets.get.return_value = RecordSet(txt_records=[])

        # Extract zone TXT record name and value
        zone1_req = SUB_DOMAIN[0]
        zone1_domain_name = zone1_req.validation_domain_name(zone1_req.domain)
        zone1_key = zone1_req.validation(zone1_req.account_key)
        # example.com is azure zone in config
        relative_record = zone1_domain_name.replace('example.com', '').strip('.')

        self.auth.perform(SUB_DOMAIN)

        # Check azure client call counts
        self.assertEqual(self.mock_client.record_sets.get.call_count, 1)
        self.assertEqual(self.mock_client.record_sets.create_or_update.call_count, 1)

        #
        zone1_call = self.mock_client.record_sets.create_or_update.call_args_list[0]
        self.assertEqual(zone1_call[1]['zone_name'], "example.com")
        self.assertEqual(zone1_call[1]['record_type'], "TXT")
        self.assertEqual(zone1_call[1]['relative_record_set_name'], relative_record)
        zone1_txt_records = zone1_call[1]['parameters'].txt_records
        self.assertEqual(len(zone1_txt_records), 1)
        self.assertEqual(zone1_txt_records[0].value[0], zone1_key)

    def test_cleanup_multiple(self):
        self.mock_client.record_sets.get.return_value = RecordSet(txt_records=[])

        # Extract zone TXT record name and value
        zone1_req, zone2_req, zone3_req = MULTI_DOMAIN
        zone1_domain_name = zone1_req.validation_domain_name(zone1_req.domain)
        zone2_domain_name = zone2_req.validation_domain_name(zone2_req.domain)
        zone3_domain_name = zone3_req.validation_domain_name(zone3_req.domain)
        zone1_relative_record = zone1_domain_name.replace('example.com', '').strip('.')
        zone2_relative_record = zone2_domain_name.replace('example.org', '').strip('.')
        zone3_relative_record = zone3_domain_name.replace('example.com', '').strip('.')

        # _attempt_cleanup | pylint: disable=protected-access
        self.auth._attempt_cleanup = True
        self.auth.cleanup(MULTI_DOMAIN)

        # Check azure client call counts
        self.assertEqual(self.mock_client.record_sets.get.call_count, 3)
        self.assertEqual(self.mock_client.record_sets.delete.call_count, 3)

        zone1_call, zone2_call, zone3_call = self.mock_client.record_sets.delete.call_args_list
        self.assertEqual(zone1_call[1]['zone_name'], "example.com")
        self.assertEqual(zone1_call[1]['record_type'], "TXT")
        self.assertEqual(zone1_call[1]['relative_record_set_name'], zone1_relative_record)

        self.assertEqual(zone2_call[1]['zone_name'], "example.org")
        self.assertEqual(zone2_call[1]['record_type'], "TXT")
        self.assertEqual(zone2_call[1]['relative_record_set_name'], zone2_relative_record)

        self.assertEqual(zone3_call[1]['zone_name'], "example.com")
        self.assertEqual(zone3_call[1]['record_type'], "TXT")
        self.assertEqual(zone3_call[1]['relative_record_set_name'], zone3_relative_record)

    def test_cleanup_existing(self):
        self.mock_client.record_sets.get.return_value = RecordSet(txt_records=[
            TxtRecord(value=['someexistingkey'])
        ])

        # Extract zone TXT record name and value
        zone1_req = SINGLE_DOMAIN[0]
        zone1_domain_name = zone1_req.validation_domain_name(zone1_req.domain)
        zone1_relative_record = zone1_domain_name.replace('example.com', '').strip('.')
        zone1_key = zone1_req.validation(zone1_req.account_key)

        # _attempt_cleanup | pylint: disable=protected-access
        self.auth._attempt_cleanup = True
        self.auth.cleanup(SINGLE_DOMAIN)

        # Check azure client call counts
        self.assertEqual(self.mock_client.record_sets.get.call_count, 1)
        self.assertEqual(self.mock_client.record_sets.delete.call_count, 0)
        self.assertEqual(self.mock_client.record_sets.create_or_update.call_count, 1)

        # Check recordset is updated to not include key
        zone1_call = self.mock_client.record_sets.create_or_update.call_args_list[0]
        self.assertEqual(zone1_call[1]['zone_name'], "example.com")
        self.assertEqual(zone1_call[1]['record_type'], "TXT")
        self.assertEqual(zone1_call[1]['relative_record_set_name'], zone1_relative_record)
        zone1_txt_records = zone1_call[1]['parameters'].txt_records

        self.assertEqual(len(zone1_txt_records), 1)
        txt_values = zone1_txt_records[0].value
        self.assertNotIn(zone1_key, txt_values)
        self.assertIn('someexistingkey', txt_values)

    def test_config_missing_auth(self):
        # Test no auth info
        dns_test_common.write({}, self.sp_config.azure_config)
        with self.assertRaises(errors.PluginError) as cm:
            self.auth.perform(SINGLE_DOMAIN)
        self.assertIn('No authentication methods have been configured', cm.exception.args[0])

    def test_config_missing_zone_info(self):
        # Test missing mapping list
        dns_test_common.write({
            'azure_sp_client_id': '912ce44a-0156-4669-ae22-c16a17d34ca5',
            'azure_sp_client_secret': 'E-xqXU83Y-jzTI6xe9fs2YC~mck3ZzUih9',
            'azure_tenant_id': 'ed1090f3-ab18-4b12-816c-599af8a88cf7',
        }, self.sp_config.azure_config)
        with self.assertRaises(errors.PluginError) as cm:
            self.auth.perform(SINGLE_DOMAIN)
        self.assertIn('At least one zone mapping needs to be provided', cm.exception.args[0])

    def test_config_bad_zone_info(self):
        # Test missing mapping list
        dns_test_common.write({
            'azure_sp_client_id': '912ce44a-0156-4669-ae22-c16a17d34ca5',
            'azure_sp_client_secret': 'E-xqXU83Y-jzTI6xe9fs2YC~mck3ZzUih9',
            'azure_tenant_id': 'ed1090f3-ab18-4b12-816c-599af8a88cf7',
            'azure_zone1': 'example.com',
        }, self.sp_config.azure_config)
        with self.assertRaises(errors.PluginError) as cm:
            self.auth.perform(SINGLE_DOMAIN)
        self.assertIn('DNS Zone mapping is not in the format', cm.exception.args[0])

    def test_config_bad_resource_group(self):
        # Test invalid resource group ID
        dns_test_common.write({
            'azure_sp_client_id': '912ce44a-0156-4669-ae22-c16a17d34ca5',
            'azure_sp_client_secret': 'E-xqXU83Y-jzTI6xe9fs2YC~mck3ZzUih9',
            'azure_tenant_id': 'ed1090f3-ab18-4b12-816c-599af8a88cf7',
            'azure_zone1': 'example.com:/subscriptions/c135abce-d87d-48df-936c-15596c6968a5/invalid',
        }, self.sp_config.azure_config)
        with self.assertRaises(errors.PluginError) as cm:
            self.auth.perform(SINGLE_DOMAIN)
        self.assertIn('Failed to parse resource ID for example.com', cm.exception.args[0])


if __name__ == "__main__":
    unittest.main()  # pragma: no cover



from setuptools import find_packages
from setuptools import setup

version = '2.5.0'

# azure-mgmt-dns is still the old style SDK, so will change dramatically
# when they refactor, most notably the credential parts
install_requires = [
    'azure-identity>=1.13.0',
    'azure-mgmt-dns>=8.0.0',
    'azure-core>=1.25.0',
    'setuptools>=41.6.0',
    'certbot>=2.0,<3.0'
]

with open("README.md") as f:
    long_description = f.read()

docs_extras = [
    'Sphinx>=1.0',  # autodoc_member_order = 'bysource', autodoc_default_flags
    'sphinx_rtd_theme',
]

setup(
    name='certbot-dns-azure',
    version=version,
    description="Azure DNS Authenticator plugin for Certbot",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/binkhq/certbot-dns-azure',
    author="Terry Cain",
    author_email='opensource@bink.com',
    license='Apache License 2.0',
    python_requires='>=3.6',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Plugins',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Security',
        'Topic :: System :: Installation/Setup',
        'Topic :: System :: Networking',
        'Topic :: System :: Systems Administration',
        'Topic :: Utilities',
    ],

    packages=find_packages(),
    include_package_data=True,
    install_requires=install_requires,
    extras_require={
        'docs': docs_extras,
    },
    entry_points={
        'certbot.plugins': [
            'dns-azure = certbot_dns_azure._internal.dns_azure:Authenticator',
        ],
    },
)
